/* ex: set tabstop=8 expandtab: */
/*
#
# MPEG2Parser class testing tool
#
# $Id: mpegcat.c,v 1.4 2002/03/11 14:53:10 nemies Exp $
#
# Copyright (C) 2001 Kees Cook
# kees@outflux.net, http://outflux.net/
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# http://www.gnu.org/copyleft/gpl.html
#
*/

#include "../config.h"

#include <stdio.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/uio.h>
#include <string.h>
#include <getopt.h>
#include "mpeg2structs.h"

int no_slices = 0;
int no_audio = 0;
int no_video = 0;
int no_pack = 0;
int no_sequence = 0;
int no_weird_streams = 0;
int no_system_headers = 0;
off_t begin_at = 0;
off_t num_bytes = 0;

static struct option long_options[] = {
	{"help", 0, 0, 'h'},
	{"slices", 0, 0, 's'},
	{"audio", 0, 0, 'a'},
	{"video", 0, 0, 'v'},
	{"pack", 0, 0, 'p'},
	{"sequence", 0, 0, 'q'},
	{"system", 0, 0, 'y'},
	{"weird", 0, 0, 'w'},
	{"just-frames", 0, 0, 'F'},
	{"begin", 0, 0, 'b'},
	{"num", 0, 0, 'n'},
	{0, 0, 0, 0}
};
static char *short_options = "h?savpqywFb:n:";

void usage(char *argv[])
{
	printf("Usage: %s [options ...] MPEG2-PS ...\n", argv[0]);
	printf("-s/--slices       Toggle slice starts\n");
	printf("-a/--audio        Toggle audio ES packet start display\n");
	printf("-v/--video        Toggle video ES packet start display\n");
	printf("-p/--pack         Toggle Pack packet start display\n");
	printf("-q/--sequence     Toggle sequence header display\n");
	printf("-y/--system       Toggle system header display\n");
	printf
	    ("-w/--weird        Toggle weird stream display (only show video & audio)\n");
	printf
	    ("-F/--just-frames  Toggle frames display (same as '-qvwaspy')\n");
	printf("-b/--begin=NUM    Start reading at offset NUM\n");
	printf("-n/--num=NUM      Stop after reading NUM bytes\n");
	printf("-h/-?/--help      Display this help\n");

	exit(1);
}

int handle_options(int argc, char *argv[])
{
	int c;
	int option_index = 0;

	while (1) {
		c = getopt_long(argc, argv, short_options, long_options,
				&option_index);
		if (c == -1)
			break;

		switch (c) {
		case '?':
		case 'h':
			usage(argv);
			break;

		case 's':
			no_slices = !no_slices;
			break;

		case 'a':
			no_audio = !no_audio;
			break;

		case 'v':
			no_video = !no_video;
			break;

		case 'p':
			no_pack = !no_pack;
			break;

		case 'q':
			no_sequence = !no_sequence;
			break;

		case 'w':
			no_weird_streams = !no_weird_streams;
			break;

		case 'y':
			no_system_headers = !no_system_headers;
			break;

		case 'F':
			no_sequence = !no_sequence;
			no_pack = !no_pack;
			no_slices = !no_slices;
			no_video = !no_video;
			no_audio = !no_audio;
			no_weird_streams = !no_weird_streams;
			no_system_headers = !no_system_headers;
			break;

		case 'n':
			num_bytes = atoll(optarg);
			break;

		case 'b':
			begin_at = atoll(optarg);
			break;

		default:
			printf
			    ("?? getopt returned character code 0x%x ??\n",
			     c);
			break;
		}
	}


	return optind;
}


#define BUFFER_SIZE	(1024*32)
FILE *fp = NULL;
uint8_t buffer[BUFFER_SIZE];
off_t offset = 0;
size_t buffer_min = 0;
size_t buffer_max = 0;

void fill_buffer()
{
	size_t got;

	//fprintf(stderr,"filling...\n");

	if (buffer_min < buffer_max) {
		memmove(buffer, buffer + buffer_min,
			buffer_max - buffer_min);
	}
	buffer_max -= buffer_min;
	buffer_min = 0;

	got =
	    fread(buffer + buffer_min, sizeof(uint8_t),
		  BUFFER_SIZE - buffer_max, fp);
	if (got < 0) {
		perror("fread");
		exit(2);
	}
	offset += got;
	buffer_max += got;
	//fprintf(stderr,"bump %d more\n",got);
}

int start_buffer(char *filename)
{
	if (fp)
		fclose(fp);
	fp = NULL;


	if (!(fp = fopen(filename, "r"))) {
		perror(filename);
		return -1;
	}

	if (begin_at) {
		if (fseek(fp, begin_at, SEEK_SET) < 0) {
			perror("fseek");
			return -2;
		}
	}

	offset = 0;
	buffer_min = 0;
	buffer_max = 0;
	fill_buffer();
	return 0;
}

int get_byte(uint8_t * byte)
{
	if (!byte)
		return -2;
	if (buffer_min == buffer_max)
		fill_buffer();
	if (buffer_min == buffer_max)
		return -1;

	*byte = buffer[buffer_min++];
	return 0;
}

int look_ahead(uint8_t * bytes, int count)
{
	int i;

	if (!bytes)
		return -2;
	if (count <= 0)
		return 0;
	if (count >= BUFFER_SIZE)
		return -3;

	if (buffer_min + count > buffer_max)
		fill_buffer();
	if (buffer_min + count > buffer_max)
		return -1;

	for (i = 0; i < count; i++) {
		*(bytes++) = buffer[buffer_min + i];
		//fprintf(stderr,"\tsaw: 0x%02X\n",buffer[buffer_min+i]);
	}

	return 0;
}

off_t buffer_tell()
{
	return begin_at + offset - buffer_max + buffer_min;
}

char *speed_str(int speed)
{
	static char answer[64];

	if (speed > 1000000) {
		sprintf(answer, "%0.2fMb/s", (float) speed / 1000000.0);
	} else if (speed > 1000) {
		sprintf(answer, "%0.2fKb/s", (float) speed / 1000.0);
	} else
		sprintf(answer, "%db/s", speed);

	return answer;
}


static char *aspect_str[] = {
	"forbidden",		// 0000
	"1:1",			// 0001
	"3:4",			// 0010
	"9:16",			// 0011
	"1:2.21",		// 0100
	"reserved",		// 0101
	"reserved",		// 0110
	"reserved",		// 0111
	"reserved",		// 1000
	"reserved",		// 1001
	"reserved",		// 1010
	"reserved",		// 1011
	"reserved",		// 1100
	"reserved",		// 1101
	"reserved",		// 1110
	"reserved",		// 1111
};
static char *frame_str[] = {
	"forbidden",		// 0000
	"23.9",			// 0001
	"24",			// 0010
	"25",			// 0011
	"29.9",			// 0100
	"30",			// 0101
	"50",			// 0110
	"59.9",			// 0111
	"60",			// 1000
	"reserved",		// 1001
	"reserved",		// 1010
	"reserved",		// 1011
	"reserved",		// 1100
	"reserved",		// 1101
	"reserved",		// 1110
	"reserved",		// 1111
};
static char *pic_str[] = {
	"forbidden",		// 000
	"I-Frame",		// 001
	"P-Frame",		// 010
	"B-Frame",		// 011
	"invalid",		// 100
	"reserved",		// 101
	"reserved",		// 110
	"reserved",		// 111
};
static char *length_format;

char *code_string(uint8_t code)
{
	static char answer[128];
	uint8_t bytes[32];

	// length variables
	// HI LO
	int length;
	static char length_str[64];

	// sequence header variables
	//|horizontal |vertical   |asp|rat|bit rate         |c
	//|   0   |   1   |   2   |   3   |   4   |   5   |   6   |
	// 76543210765432107654321076543210765432107654321076543210
	int horizontal_size;
	int vertical_size;
	int aspect_ratio;
	int frame_rate;
	int bit_rate;
	int constrained;

	// picture header variables;
	//|temporal |ty|
	//|   0   |   1   |   2   |   3   |   4   |   5   |   6   |
	// 76543210765432107654321076543210765432107654321076543210
	int pic_type;
	int pic_time;

	// GOP header variables
	// |    0    |    1    |    2   |    3    |
	//  7 65432 107654 3 210765 432107 6 543210
	//  1 11111 111111 1 111111 111111 1 1
	//  d hour  min    m sec    pic    c b
	//  r              a               l roken
	//  o              r               osed
	//  p              k
	int drop;
	int hour;
	int min;
	int sec;
	int pictures;
	int closed;
	int broken;

#define CALC_LENGTH_STR	\
	{	\
		length=0;	\
		if (look_ahead(bytes,2)==0)	\
			length=bytes[1] | (bytes[0] << 8);	\
		sprintf(length_str,length_format,	\
			length,length==1?"":"s",buffer_tell()+length+2);\
	}

	switch (code) {
	case PES_TYPE_picture_start:
		if (look_ahead(bytes, 2) == 0) {
			pic_type = (bytes[1] & PICTURE_CODING_MASK) >> 3;
			pic_time =
			    (bytes[0] << 2) | ((bytes[1] & 0xC0) >> 6);

			sprintf(answer, "  Picture (%s) Start (%u)",
				pic_str[pic_type], pic_time);
			return answer;
		} else
			return "  Picture Start (incomplete flags)";
	case PES_TYPE_user_data_start:
		return " User Data Start";
	case PES_TYPE_sequence_header:
		if (no_sequence)
			return NULL;

		if (look_ahead(bytes, 7) == 0) {
			horizontal_size =
			    (bytes[0] << 4) | ((bytes[1] & 0xF0) >> 4);
			vertical_size =
			    ((bytes[1] & 0x0F) << 8) | bytes[2];
			aspect_ratio = (bytes[3] & 0xF0) >> 4;
			frame_rate = (bytes[3] & 0x0F);
			bit_rate =
			    (bytes[4] << 10) | (bytes[5] << 2) |
			    ((bytes[6] & 0xC0) >> 6);
			constrained = (bytes[6] & 0x20) >> 5;

			sprintf(answer,
				"  Sequence Header (%dx%d: %s, %sfps, %s%s)",
				horizontal_size, vertical_size,
				aspect_str[aspect_ratio],
				frame_str[frame_rate],
				speed_str(bit_rate * 400),
				constrained ? " constrained" : "");
			return answer;
		} else
			return "  Sequence Header (incomplete flags)";
	case PES_TYPE_sequence_error:
		if (no_sequence)
			return NULL;
		return "  Sequence Error";
	case PES_TYPE_extension_start:
		if (no_sequence)
			return NULL;
		return "  Sequence Extension Start";
	case PES_TYPE_sequence_end:
		if (no_sequence)
			return NULL;
		return "  Sequence End";
	case PES_TYPE_group_start:

		if (look_ahead(bytes, 4) == 0) {
			drop = ((bytes[0] & 0x80) > 0);
			hour = ((bytes[0] & 0x7C) >> 2);
			min =
			    ((bytes[0] & 0x3) << 4) | ((bytes[1] & 0xF0) >>
						       4);
			sec =
			    ((bytes[1] & 0x7) << 3) | ((bytes[2] & 0xE0) >>
						       6);
			pictures =
			    ((bytes[2] & 0x3F) << 1) | ((bytes[3] & 0x80)
							>> 7);
			closed = ((bytes[3] & 0x40) > 0);
			broken = ((bytes[3] & 0x20) > 0);

			sprintf(answer,
				"  GOP Start (%02d:%02d:%02d.%02d%s%s%s)",
				hour, min, sec, pictures,
				drop ? " drop" : "",
				closed ? " closed" : "",
				broken ? " broken" : "");
			return answer;
		} else
			return "  GOP Start (incomplete flags)";
	case PES_TYPE_program_end:
		return "Program End";
	case PES_TYPE_pack_start:
		if (no_pack)
			return NULL;
		return "Pack Start";
	case PES_TYPE_system_header:
		if (no_system_headers)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " System Header: %s", length_str);
		return answer;
	case PES_TYPE_program_stream_map:
		return " Program Stream Map";
		// FIXME: put all the weird streams into one section under default
	case PES_TYPE_private_stream_1:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " Private Stream 1: %s", length_str);
		return answer;
	case PES_TYPE_padding_stream:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " Padding Stream: %s", length_str);
		return answer;
	case PES_TYPE_private_stream_2:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " Private Stream 2: %s", length_str);
		return answer;
	case PES_TYPE_ECM_stream:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " ECM Stream: %s", length_str);
		return answer;
	case PES_TYPE_EMM_stream:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " EMM Stream: %s", length_str);
		return answer;
	case PES_TYPE_DSMCC_stream:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " DSMCC Stream: %s", length_str);
		return answer;
	case PES_TYPE_ISO_13522:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " ISO13522 Stream: %s", length_str);
		return answer;
	case PES_TYPE_H2221A:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " H222.1A: %s", length_str);
		return answer;
	case PES_TYPE_H2221B:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " H222.1B: %s", length_str);
		return answer;
	case PES_TYPE_H2221C:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " H222.1C: %s", length_str);
		return answer;
	case PES_TYPE_H2221D:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " H222.1D: %s", length_str);
		return answer;
	case PES_TYPE_H2221E:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " H222.1E Stream: %s", length_str);
		return answer;
	case PES_TYPE_ancillary_stream:
		if (no_weird_streams)
			return NULL;
		CALC_LENGTH_STR;
		sprintf(answer, " Ancillary Stream: %s", length_str);
		return answer;
	case PES_TYPE_program_stream_directory:
		return " Program Stream Directory";
	default:
		if (code >= 0x01 && code <= 0xAF) {
			if (no_slices)
				return NULL;
			return "  Slice Start";
		}
		if ((code & PES_TYPE_MASK_video) == PES_TYPE_video) {
			if (no_video)
				return NULL;

			CALC_LENGTH_STR;

			sprintf(answer, " Video Stream 0x%02X: %s",
				PES_TYPE_video -
				(code & PES_TYPE_MASK_video), length_str);
			return answer;
		}
		if ((code & PES_TYPE_MASK_audio) == PES_TYPE_audio) {
			if (no_audio)
				return NULL;

			CALC_LENGTH_STR;

			sprintf(answer, " Audio Stream 0x%02X: %s",
				PES_TYPE_audio -
				(code & PES_TYPE_MASK_audio), length_str);
			return answer;
		}
		if ((code >= 0xFA && code <= 0xFE) ||
		    (code >= 0xB0 && code <= 0xB1) || (code == 0xB6)) {
			return "Reserved";
		}


		return "Unknown";

		break;
	}
	return "ERROR: no match?!";
}


int main(int argc, char *argv[])
{
	int i, optind;
	uint32_t marker;
	uint8_t byte;
	char *report_format;

	optind = handle_options(argc, argv);

	if (optind >= argc)
		usage(argv);

	if (sizeof(off_t) > sizeof(long)) {
		report_format = "%10llu: 0x%02X: %s\n";
		length_format = "%d byte%s (next packet @ %llu)";
	} else {
		report_format = "%10lu: 0x%02X: %s\n";
		length_format = "%d byte%s (next packet @ %lu)";
	}

	for (i = optind; i < argc; i++) {
		int running = 1;

		if (start_buffer(argv[i]) < 0)
			continue;

		marker = 0xFFFFFFFF;
		while (running) {
			/*
			   fprintf(stderr,"marker: 0x%02X 0x%02X 0x%02X 0x%02X\n",
			   (marker & 0xFF000000) >> 24,
			   (marker & 0x00FF0000) >> 16,
			   (marker & 0x0000FF00) >> 8,
			   (marker & 0x000000ff));
			 */
			if ((marker & 0xFFFFFF00) == 0x100) {
				char *str;

				if ((str = code_string(byte))) {
					printf(report_format,
					       buffer_tell() - 4, byte,
					       str);
				}
			}
			marker <<= 8;
			if (get_byte(&byte) < 0) {
				printf(report_format, buffer_tell(), 0x00,
				       "EOF");
				running = 0;
			} else
				marker |= byte;

			if (num_bytes
			    && buffer_tell() - begin_at > num_bytes) {
				printf(report_format, buffer_tell(), 0x00,
				       "EOB");
				running = 0;
			}
		}
	}

	return 0;
}
